#!/usr/bin/python3
import os
import sys
import time
import random
import argparse
import subprocess
from threading import Thread
from datetime import datetime, timedelta
from dataEngine import Aws, Ami, Conversation, Channel, Socket, AppConfig, AmiMock, PcapReader, DumpWriter, Logger, EventHandlers, Ldap

logger = Logger.get()
print = logger.debug

# TODO
# 1. Use environment variable for aws_access_key_id and aws_secret_access_key
# 2. add queue to push messages to kinesis, send it by batch to reduce payments
# 3. Add systemd service for use this script
# 4. Add all python's dependencies (boto3, asterisk.ami, ldap3) to deb package
# 5. Add cache for LDAP queries
# 6. Add BRIDGE_START logic for app queue (Local channels)

# CHECK
# 1. All kind of transfers
#   1.1. Hardware transfer (blind / attendant / semy attendant transfer) -- DONE
#   1.2. Collaboraation transfer (blind / attendant / semy attendant transfer) -- DONE
#   1.3. Feature сode (blind / attendant / semy attendant transfer) -- DONE
# 2. REFER after DirectRTP -- DONE
# 3. Pickup -- DONE
# 4. Forward -- DONE
# 5. Redirect -- DONE
#####


def update_active_calls():
    while True:
        time.sleep(30)
        for channel in Conversation.get_conversations_list():
            if datetime.now() - Conversation.get(channel)["event"]["time"] > timedelta(minutes=5):
                Conversation.get(channel).call_live()


def stop(delay=None):
    msg = " with a delay of %d seconds" % delay if delay else ""
    print("Stop{}".format(msg))

    PcapReader.close()
    Ami.close_connection()
    Socket.close_connection()
    AmiMock.close_connection()
    DumpWriter.close()
    Ldap.store()

    if delay:
        time.sleep(delay)
    sys.exit()


def get_free_port():
    out = -1
    while out != 0:
        port = random.randint(7000, 65000)
        p1 = subprocess.Popen(["netstat", "-npa"], stdout=subprocess.PIPE)
        p2 = subprocess.Popen(["grep", str(port)], stdin=p1.stdout, stdout=subprocess.PIPE)
        p3 = subprocess.Popen(["wc", "-l"], stdin=p2.stdout, stdout=subprocess.PIPE)
        (out, err) = p3.communicate()
        out = int(out.decode("utf-8"))
    return port


def parse_args(argv):
    def str_to_bool(value):
        if isinstance(value, bool):
            return value
        if value.lower() in ('false', 'f', '0', 'no', 'n'):
            return False
        elif value.lower() in ('true', 't', '1', 'yes', 'y'):
            return True
        raise ValueError(f'{value} is not a valid boolean value')

    def str_to_loglevel(value):
        if value.isdigit():
            value = int(value)
            if value == 0:
                return "ERROR"
            if value == 1:
                return "WARNING"
            if value == 2:
                return "INFO"
            if value >= 3:
                return "DEBUG"
        elif value.upper() in ("ERROR", "WARNING", "INFO", "DEBUG"):
            return value.upper()
        raise ValueError(f'{value} is not a valid logger value')

    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--in_file', default=None, nargs='+',
                        help='path to pcap input file collected by `tcpdump -s0 -ilo port $(netstat -npa | grep 5039 | grep python | cut -d: -f2 | cut -d' ' -f1) -w IN_FILE` or start script with parameter --dump to collect dump file. '
                             'It doen\'t work with paramater --dump')
    parser.add_argument('-o', '--out_file', help='path to output file with json events', default=None)
    parser.add_argument('-rwo', '--rewrite_out_file', nargs='?', type=str_to_bool, const=True, default=None,
                        help='rewrite output file')
    parser.add_argument('-dp', '--disable_push', help='disable push events to the Kinesis data stream',
                        type=str_to_bool, nargs='?', default=None, const=True)
    parser.add_argument('-p', '--port', help='port of pseudo AMI interface', type=int, default=None)
    parser.add_argument('-l', '--loglevel', help='log level (0=ERROR | 1=WARNING | 2=INFO | 3=DEBUG)', type=str_to_loglevel, default=None)
    parser.add_argument('-cl', '--cache_ldap', help='cache ldap results data', type=str_to_bool, default=None, const=True, nargs='?')
    parser.add_argument('-fd', '--full_debug', help='enagle full debug logging (include all external liraries)',
                        type=str_to_bool, default=False, const=True, nargs='?')
    parser.add_argument('-e', '--show_events', help='display all ami events (log level has to be set to debug)',
                        type=str_to_bool, default=None, const=True, nargs='?')
    parser.add_argument('-rt', '--realtime', help='use realtime mode with timeout for test mode (parameter --in_file)',
                        nargs='?', const=True, type=str_to_bool, default=False)
    parser.add_argument('--dump', nargs='?', help='Name of test case [default: `testcase`]. Write tcpdump of AMI communication (testcase.pcap) and events to json file (testcase.json). It doen\'t work with paramater --in_file. Must be started with root privilege',
                        default=None, const='testcase')
    args = parser.parse_args()
    print("args: {}".format(args))

    AppConfig.set("cache_ldap", args.cache_ldap)

    if args.in_file:
        AppConfig.set("in_file", args.in_file)
        port = get_free_port()
        AppConfig.set("port", int(port))
        AppConfig.set("isTest", True)
        AppConfig.set("disable_push", True)
        AppConfig.set("rewrite_out_file", True)
        AppConfig.set("realtime", args.realtime)
    elif args.dump:
        AppConfig.set("rewrite_out_file", True)
        AppConfig.set("out_file", "{}.json".format(args.dump))
        AppConfig.set("dump", "{}.pcap".format(args.dump))
        AppConfig.set("ldap_file", "{}.ldap".format(args.dump))
        AppConfig.set("cache_ldap", True)

    if args.loglevel:
        logger.setLevel(args.loglevel)

    AppConfig.set("out_file", args.out_file)
    AppConfig.set("disable_push", args.disable_push)
    AppConfig.set("port", args.port)
    AppConfig.set("rewrite_out_file", args.rewrite_out_file)
    if args.show_events:
        AppConfig.set("show_events", args.show_events)
        logger.setLevel("DEBUG")
    AppConfig.set("full_debug", args.full_debug)


def cleanup_all_data():
    Channel.end_all()
    Conversation.end_all()


def main(argv):
    parse_args(argv)
    Logger.init_logger()
    logger.info("Starting")

    Socket.init_connection()
    if not Aws.check_cognito_credentials():
        stop(3600)
    Aws.init_client()

    call_live = Thread(target=update_active_calls)
    call_live.daemon = True
    call_live.start()

    if AppConfig.get("rewrite_out_file") and AppConfig.get("out_file") and os.path.isfile(AppConfig.get("out_file")):
        os.remove(AppConfig.get("out_file"))

    AmiMock.init_connection()
    Ami.init_connection(EventHandlers.event_listener, cleanup_all_data)
    PcapReader.init()
    DumpWriter.init()

    try:
        while True:
            # Channel.print_current_state()
            # Conversation.print_current_state()
            if not Aws.check_cognito_credentials():
                stop(3600)
            if AppConfig.isStopped():
                break
            time.sleep(5)
    except (KeyboardInterrupt, SystemExit):
        pass

    stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))
